/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __MODEL__
#define __MODEL__

#include "modelgen.h"
#include "spritegn.h"

/*

d*_t structures are on-disk representations
m*_t structures are in-memory

*/

// entity effects

#define	EF_BRIGHTFIELD			1
#define	EF_MUZZLEFLASH 			2
#define	EF_BRIGHTLIGHT 			4
#define	EF_DIMLIGHT 			8


/*
==============================================================================

BRUSH MODELS

In memory representations expand short to int

==============================================================================
*/

// 
// in memory representation
// 
// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct
{
	float		position[3];
	float		dlrgb[3];
	int			dlightframe;
} mvertex_t;

typedef struct
{
	vec3_t		dir;
} mnormal_t;

typedef struct
{
	vec4_t		dir;
} mtangent_t;

typedef struct
{
	vec2_t	st;
} mtexcoord_t;

#define	SIDE_FRONT	0
#define	SIDE_BACK	1
#define	SIDE_ON		2


// plane_t structure
// !!! if this is changed, it must be changed in asm_i386.h too !!!
typedef struct mplane_s
{
	vec3_t	normal;
	float	dist;
	byte	type;			// for texture axis selection and fast side tests
	byte	signbits;		// signx + signy<<1 + signz<<1
	byte	pad[2];
	int		cacheframe;
	entity_t *cacheent;
	float cachedist;
} mplane_t;

typedef struct texture_s
{
	char				name[16];
	unsigned			width, height;
	struct gltexture_s	*gltexture; // johnfitz -- pointer to gltexture
	struct gltexture_s	*fullbright; // johnfitz -- fullbright mask texture
	struct gltexture_s	*normalmap;	// mh - normal map texture
	int					anim_total;				// total tenths in sequence ( 0 = no)
	int					anim_min, anim_max;		// time for this frame min <=time< max
	struct texture_s	*anim_next;		// in the animation sequence
	struct texture_s	*alternate_anims;	// bmodels in frmae 1 use these

	struct msurface_s	*texturechain;
} texture_t;


#define	SURF_PLANEBACK		2
#define	SURF_DRAWSKY		4
#define SURF_DRAWSPRITE		8
#define SURF_DRAWTURB		0x10
#define SURF_DRAWTILED		0x20
#define SURF_DRAWBACKGROUND	0x40
#define SURF_UNDERWATER		0x80
#define SURF_NOTEXTURE		0x100 // johnfitz
#define SURF_DRAWFENCE		0x200
#define SURF_DRAWLAVA		0x400
#define SURF_DRAWSLIME		0x800
#define SURF_DRAWTELE		0x1000
#define SURF_DRAWWATER		0x2000

// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct
{
	unsigned int	v[2];
	unsigned int	cachededgeoffset;
} medge_t;

typedef struct
{
	float		vecs[2][4];
	float		mipadjust;
	texture_t	*texture;
	int			flags;
} mtexinfo_t;


typedef struct glvertex_s
{
	union
	{
		struct
		{
			float v[3];
			float st1[2];
			float st2[2];
		};

		float verts[7];
	};
} glvertex_t;


// extra data for GPU lighting
typedef struct r_surfextra_s
{
	float styles[4];
	float dynamic[3];
} r_surfextra_t;


typedef struct msurface_s
{
	int			surfnum;
	int			visframe;		// should be drawn when node is crossed
	int			clipflags;
	float		mins[3];		// johnfitz -- for frustum culling
	float		maxs[3];		// johnfitz -- for frustum culling

	// basic position verts for a surf
	mvertex_t	**baseverts;
	float		*matrix;
	float		midpoint[3];
	float		dist;

	struct msurface_s	*texturechain;
	struct msurface_s	*lightmapchain;

	mplane_t	*plane;
	int			flags;

	int			firstedge;	// look up in model->surfedges[], negative numbers
	int			numedges;	// are backwards edges

	int			streamoffset;
	int			firstvertex;
	int			vertexrange;
	int			numvertexes;
	int			numindexes;

	// the model that uses this surface (for vertex regeneration)
	struct model_s		*model;

	int		texturemins[2];
	int		extents[2];

	// mh - lightmaps
	int			smax;
	int			tmax;
	int			light_s;
	int			light_t;

	mtexinfo_t	*texinfo;

	// explicit water alpha used by this surf
	float		wateralpha;

	// lighting info
	int			dlightframe;
	int			dlightbits[4];	// mh - 128 dynamic lights

	int			lightmaptexturenum;

	GLuint		vertexprogram;
	GLuint		fragmentprogram;

	byte		styles[MAX_SURFACE_STYLES];
	float		cached_light[MAX_SURFACE_STYLES];
	qboolean	cached_dlight;
	int			numstyles;
	struct r_surfextra_s *extra;
	byte		*samples;		// [numstyles*surfsize]
} msurface_t;


typedef struct mnode_s
{
	// common with leaf
	int			contents;		// 0, to differentiate from leafs
	int			visframe;		// node needs to be traversed if current

	float		minmaxs[6];		// for bounding box culling

	struct mnode_s	*parent;

	// node specific
	mplane_t	*plane;
	struct mnode_s	*children[2];

	unsigned int		firstsurface;
	unsigned int		numsurfaces;
} mnode_t;



typedef struct mleaf_s
{
	// common with node
	int			contents;		// wil be a negative contents number
	int			visframe;		// node needs to be traversed if current

	float		minmaxs[6];		// for bounding box culling

	struct mnode_s	*parent;

	// leaf specific
	byte		*compressed_vis;
	efrag_t		*efrags;

	msurface_t	**firstmarksurface;
	int			nummarksurfaces;
	int			key;			// BSP sequence number for leaf's contents
	byte		ambient_sound_level[NUM_AMBIENTS];
} mleaf_t;

// johnfitz -- for clipnodes>32k
typedef struct mclipnode_s
{
	int			planenum;
	int			children[2]; // negative numbers are contents
} mclipnode_t;
// johnfitz

// !!! if this is changed, it must be changed in asm_i386.h too !!!
typedef struct
{
	mclipnode_t	*clipnodes; // johnfitz -- was dclipnode_t
	mplane_t	*planes;
	int			firstclipnode;
	int			lastclipnode;
	vec3_t		clip_mins;
	vec3_t		clip_maxs;
} hull_t;

/*
==============================================================================

SPRITE MODELS

==============================================================================
*/


// FIXME: shorten these?
typedef struct mspriteframe_s
{
	int					width, height;
	float				up, down, left, right;
	float				smax, tmax; // johnfitz -- image might be padded
	struct gltexture_s	*gltexture;
} mspriteframe_t;

typedef struct
{
	int				numframes;
	float			*intervals;
	mspriteframe_t	*frames[1];
} mspritegroup_t;

typedef struct
{
	spriteframetype_t	type;
	mspriteframe_t		*frameptr;
} mspriteframedesc_t;

typedef struct
{
	int					type;
	int					maxwidth;
	int					maxheight;
	int					numframes;
	qboolean			litsprite;
	float				beamlength;		// remove?
	void				*cachespot;		// remove?
	mspriteframedesc_t	frames[1];
} msprite_t;


/*
==============================================================================

ALIAS MODELS

Alias models are position independent, so the cache manager can move them.
==============================================================================
*/

typedef struct aliasbbox_s
{
	float				mins[3];
	float				maxs[3];
} aliasbbox_t;

typedef struct
{
	int					firstpose;
	int					numposes;
	float				interval;

	int					frame;
	char				name[16];
} maliasframedesc_t;

typedef struct
{
	int					frame;
} maliasgroupframedesc_t;

typedef struct
{
	int						numframes;
	int						intervals;
	maliasgroupframedesc_t	frames[1];
} maliasgroup_t;

// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct mtriangle_s
{
	int					facesfront;
	int					vertindex[3];
} mtriangle_t;


#define	MAX_SKINS	32

// split out to keep vertex sizes down
typedef struct aliasmesh_s
{
	float st[2];
	unsigned short vertindex;
} aliasmesh_t;

typedef struct meshxyz_s
{
	byte xyz[4];
	float normal[3];
} meshxyz_t;


typedef struct meshst_s
{
	float st[2];
} meshst_t;


typedef struct aliashdr_s
{
	vec3_t      scale;
	vec3_t      scale_origin;

	float      boundingradius;
	synctype_t   synctype;
	int         flags;
	float      size;

	int         nummeshframes;
	int         numtris;
	int         numverts;

	int         vertsperframe;
	int         numframes;

	int         meshverts;
	intptr_t	aliasmesh;
	intptr_t    vertexes;
	int         framevertexsize;

	// only shaders use these
	intptr_t    vertindexes;
	qboolean	delerped;
	intptr_t	lerptable;

	int         skinwidth;
	int         skinheight;
	int         numskins;

	intptr_t	meshdesc;
	intptr_t	indexes;
	int			numindexes;

	intptr_t	firstindex;
	intptr_t	firstvertex;

	intptr_t	bboxes;

	// only vbos use these
	int			vboindexofs;
	int			vboxyzofs;
	int			vbostofs;

	struct gltexture_s	*gltextures[MAX_SKINS][4]; // johnfitz
	struct gltexture_s	*fbtextures[MAX_SKINS][4]; // johnfitz
	int               texels[MAX_SKINS];   // only for player skins
	maliasframedesc_t   frames[1];
} aliashdr_t;


#define	MAXALIASVERTS	2000 // johnfitz -- was 1024
#define	MAXALIASFRAMES	256
#define	MAXALIASTRIS	2048

extern	aliashdr_t	*pheader;
extern	stvert_t	stverts[MAXALIASVERTS];
extern	mtriangle_t	triangles[MAXALIASTRIS];
extern	trivertx_t	*poseverts[MAXALIASFRAMES];

// ===================================================================

// 
// Whole model
// 

typedef enum {mod_brush, mod_sprite, mod_alias, mod_iqm} modtype_t;

#define	EF_ROCKET	1			// leave a trail
#define	EF_GRENADE	2			// leave a trail
#define	EF_GIB		4			// leave a trail
#define	EF_ROTATE	8			// rotate (bonus items)
#define	EF_TRACER	16			// green split trail
#define	EF_ZOMGIB	32			// small blood trail
#define	EF_TRACER2	64			// orange split trail + rotate
#define	EF_TRACER3	128			// purple trail

// johnfitz -- extra flags for rendering
// (mh - changed these to higher numbers so that we can add new EF_ flags to MDLs more easily and cleanly)
#define	MOD_NOLERP		(1 << 24)		// don't lerp when animating
#define	MOD_NOSHADOW	(1 << 25)		// don't cast a shadow
#define	MOD_FBRIGHTHACK	(1 << 26)		// when fullbrights are disabled, use a hack to render this model brighter
// johnfitz

typedef struct model_s
{
	char		name[MAX_QPATH];
	qboolean	needload;		// bmodels and sprites don't cache normally

	modtype_t	type;
	int			numframes;
	synctype_t	synctype;

	int			flags;

	// volume occupied by the model graphics
	// rmins/rmaxs and ymins/ymaxs were crap and have been removed; we have *proper* rotating bbox support now ;)
	vec3_t		mins, maxs;

	// solid volume for clipping
	qboolean	clipbox;
	vec3_t		clipmins, clipmaxs;

	// brush model
	int			firstmodelsurface, nummodelsurfaces;

	int			numsubmodels;
	dmodel_t	*submodels;

	int			numplanes;
	mplane_t	*planes;

	int			numleafs;		// number of visible leafs, not counting 0
	mleaf_t		*leafs;

	int			numvertexes;
	mvertex_t	*vertexes;

	int			numedges;
	medge_t		*edges;

	int			numnodes;
	mnode_t		*nodes;

	int			numtexinfo;
	mtexinfo_t	*texinfo;

	int			numsurfaces;
	msurface_t	*surfaces;

	int			numsurfedges;
	int			*surfedges;

	int			numclipnodes;
	mclipnode_t	*clipnodes; // johnfitz -- was dclipnode_t

	int			nummarksurfaces;
	msurface_t	**marksurfaces;

	hull_t		hulls[MAX_MAP_HULLS];

	int			numtextures;
	texture_t	**textures;

	byte		*visdata;
	byte		*lightdata;
	char		*entities;

	// bbox for the model if it's a brush model
	vec3_t					bbmins;
	vec3_t					bbmaxs;

	struct iqmdata_s *iqmdata;

	// additional model data
	// !!!!! DO NOT place any struct members after this as the extra data may immediately follow it in memory !!!!!
	cache_user_t	cache;		// only access through Mod_Extradata
} model_t;

typedef struct iqmvertex_s
{
	float position[3];
	float normal[3];
	float texcoord[2];
} iqmvertex_t;

// move to model.h or modelgen.h when done...
typedef struct iqmdata_s
{
	int version;

	float *blendweights;	// suck it down, memory saving weenies!
	unsigned char *blendindexes;

	matrix3x4_t *baseframe;
	matrix3x4_t *frames;
	matrix3x4_t *outframe;

	union
	{
		iqmjointv1_t *jointsv1;
		iqmjointv2_t *jointsv2;
	};

	int num_joints;

	vec3_t mins;
	vec3_t maxs;
	float radius;

	// interleave vertexes for better performance on the GPU.
	// it's 2011, we have hardware vertex processing now.
	iqmvertex_t *verts;
	iqmtriangle_t *tris;

	int num_frames;
	int num_poses;
	int num_triangles;
	int numvertexes;

	int cachedlastframe;
	int cachedcurrframe;
	float cachedlerp;

	// only vbos get this
	int vertexoffset;

	iqmmesh_t *mesh;

	// only vbos get this
	int *indexoffsets;

	int num_mesh;

	struct gltexture_s **skins;
	struct gltexture_s **fullbrights;
} iqmdata_t;


// ============================================================================

void	Mod_Init (void);
void	Mod_ClearAll (void);
model_t *Mod_ForName (char *name, qboolean crash);
void	*Mod_Extradata (model_t *mod);	// handles caching
void	Mod_TouchModel (char *name);

mleaf_t *Mod_PointInLeaf (float *p, model_t *model);
byte	*Mod_LeafPVS (mleaf_t *leaf, model_t *model);

void Mod_SetExtraFlags (model_t *mod);

void Matrix3x4_TransformNormal (mnormal_t *out, matrix3x4_t mat, const mnormal_t in);
void Matrix3x4_TransformTangent (mtangent_t *out, matrix3x4_t mat, const mtangent_t in);
void Matrix3x4_Invert (matrix3x4_t *out, matrix3x4_t in);
void Matrix3x4_FromQuatAndVectors (matrix3x4_t *out, vec4_t rot, const float trans[3], const float scale[3]);
void Matrix3x4_Multiply (matrix3x4_t *out, matrix3x4_t mat1, matrix3x4_t mat2);
void Matrix3x4_Scale (matrix3x4_t *out, matrix3x4_t *in, float scale);
void Matrix3x4_Add (matrix3x4_t *out, matrix3x4_t *mat1, matrix3x4_t *mat2);
void Matrix3x4_Subtract (matrix3x4_t *out, matrix3x4_t mat1, matrix3x4_t mat2);
void Matrix3x4_Copy (matrix3x4_t *out, matrix3x4_t *in);
void Matrix3x4_Transform (mvertex_t *out, matrix3x4_t mat, const mvertex_t in);
void Matrix3x4GenRotate (matrix3x4_t *out, float angle, const vec3_t axis);
void Matrix3x4ForEntity (matrix3x4_t *out, entity_t *ent, float z);
double degreeToRadian (double degree);

#endif	// __MODEL__
