/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_misc.c

#include "quakedef.h"

GLuint current_vp = 0;
GLuint current_fp = 0;

void GL_SetVertexProgram (GLuint vp)
{
	if (current_vp != vp)
	{
		qglBindProgramARB (GL_VERTEX_PROGRAM_ARB, vp);
		current_vp = vp;
	}
}


extern qboolean showtris;
extern GLuint arb_showtris_fp;

void GL_SetFragmentProgram (GLuint fp)
{
	if (showtris) fp = arb_showtris_fp;

	if (current_fp != fp)
	{
		qglBindProgramARB (GL_FRAGMENT_PROGRAM_ARB, fp);
		current_fp = fp;
	}
}


void GL_SetPrograms (GLuint vp, GLuint fp)
{
	GL_SetVertexProgram (vp);
	GL_SetFragmentProgram (fp);
}


GLuint gl_arb_programs[1024];
int gl_num_arb_programs = 0;


GLuint GL_CreateProgram (GLenum mode, const char *source, qboolean fog)
{
	GLuint progid;
	GLint errPos;
	const GLubyte *errString;
	GLenum errGLErr;
	char *realsource = (char *) scratchbuf;

	if (mode == GL_FRAGMENT_PROGRAM_ARB && fog)
	{
		strcpy (realsource, "!!ARBfp1.0\nOPTION ARB_fog_exp2;\n");
		strcat (realsource, source + 11);
	}
	else strcpy (realsource, source);

	qglGenProgramsARB (1, &progid);
	qglBindProgramARB (mode, progid);

	errGLErr = glGetError ();
	qglProgramStringARB (mode, GL_PROGRAM_FORMAT_ASCII_ARB, strlen (realsource), realsource);
	errGLErr = glGetError ();

	// Find the error position
	glGetIntegerv (GL_PROGRAM_ERROR_POSITION_ARB, &errPos);
	errString = glGetString (GL_PROGRAM_ERROR_STRING_ARB);

	if (errGLErr != GL_NO_ERROR) Con_Printf ("Generic OpenGL Error\n");
	if (errPos != -1) Con_Printf ("Program error at position: %d\n", errPos);
	if (errString && errString[0]) Con_Printf ("Program error: %s\n", errString);

	if ((errPos != -1) || (errString && errString[0]) || (errGLErr != GL_NO_ERROR))
	{
		Con_Printf ("Program:\n%s\n", realsource);
		qglDeleteProgramsARB (1, &progid);
		qglBindProgramARB (mode, 0);
		return 0;
	}
	else
	{
		gl_arb_programs[gl_num_arb_programs] = progid;
		gl_num_arb_programs++;

		qglBindProgramARB (mode, 0);
		return progid;
	}
}


void R_DeleteShaders (void)
{
	int i;

	for (i = 0; i < gl_num_arb_programs; i++)
	{
		qglDeleteProgramsARB (1, &gl_arb_programs[i]);
		gl_arb_programs[i] = 0;
	}

	gl_num_arb_programs = 0;
}


void GLARB_SetupAliasModel (entity_t *e);
void GLARB_DrawAliasBatches  (entity_t **ents, int numents);

int r_registrationsequence = 0;

cvar_t r_showtris = {"r_showtris", "0"};

// johnfitz -- new cvars
extern cvar_t r_clearcolor;
extern cvar_t r_drawflat;
extern cvar_t r_flatlightstyles;
extern cvar_t gl_fullbrights;
extern cvar_t r_waterwarp;
extern cvar_t r_oldskyleaf;
extern cvar_t r_drawworld;
extern cvar_t r_showbboxes;
extern cvar_t r_lerpmodels;
extern cvar_t r_lerpmove;
extern cvar_t r_nolerp_list;
// johnfitz

extern cvar_t r_lockpvs;
extern cvar_t r_lockfrustum;
extern cvar_t r_colorfilter;
extern cvar_t r_normalmapping;

extern gltexture_t *playertextures[MAX_SCOREBOARD]; // johnfitz

void R_NoLerpList_f (void); // johnfitz


cvar_t gl_src_blend = {"gl_src_blend", "0"};
cvar_t gl_dst_blend = {"gl_dst_blend", "0"};


void GL_BlendTest (void)
{
	GLenum sfactors[] = {GL_ZERO, GL_ONE, GL_DST_COLOR, GL_ONE_MINUS_DST_COLOR, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_DST_ALPHA, GL_ONE_MINUS_DST_ALPHA, GL_SRC_ALPHA_SATURATE};
	GLenum dfactors[] = {GL_ZERO, GL_ONE, GL_SRC_COLOR, GL_ONE_MINUS_SRC_COLOR, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_DST_ALPHA, GL_ONE_MINUS_DST_ALPHA};

	if (gl_src_blend.value < 0) Cvar_SetValue ("gl_src_blend", 0);
	if (gl_src_blend.value > 8) Cvar_SetValue ("gl_src_blend", 8);

	if (gl_dst_blend.value < 0) Cvar_SetValue ("gl_dst_blend", 0);
	if (gl_dst_blend.value > 7) Cvar_SetValue ("gl_dst_blend", 7);

	glBlendFunc (sfactors[(int) gl_src_blend.value], dfactors[(int) gl_dst_blend.value]);
}


gltexture_t *r_whitetexture = NULL;
gltexture_t *r_greytexture = NULL;
gltexture_t *r_blacktexture = NULL;

#define ALPHATYPE_ALIAS		1
#define ALPHATYPE_BRUSH		2
#define ALPHATYPE_SPRITE	3
#define ALPHATYPE_LIQUID	4
#define ALPHATYPE_PARTICLE	5
#define ALPHATYPE_FENCE		6
#define ALPHATYPE_SURFACE	7
#define ALPHATYPE_CORONA	8
#define ALPHATYPE_IQM		9

typedef struct alphalist_s
{
	entity_t *ent;
	msurface_t *surf;
	particle_type_t *particles;
	dlight_t *light;
	float surfalpha;

	xcommand_t setupfunc;
	xcommand_t takedownfunc;

	int type;
	float dist;
} alphalist_t;


#define MAX_ALPHALIST 65536

alphalist_t *r_alphalist[MAX_ALPHALIST] = {NULL};
int r_numalphalist = 0;

float R_GetDist (float *origin)
{
	// no need to sqrt these as all we're concerned about is relative distances
	// (if x < y then sqrt (x) is also < sqrt (y))
	return
	(
		(origin[0] - r_origin[0]) * (origin[0] - r_origin[0]) +
		(origin[1] - r_origin[1]) * (origin[1] - r_origin[1]) +
		(origin[2] - r_origin[2]) * (origin[2] - r_origin[2])
	);
}


void R_AlphaListBeginMap (void)
{
	int i;

	// if we ever move this from Hunk memory we'll need to free it too...
	for (i = 0; i < MAX_ALPHALIST; i++)
		r_alphalist[i] = NULL;

	r_numalphalist = 0;
}


void R_AlphaListBeginFrame (void)
{
	r_numalphalist = 0;
}


alphalist_t *R_GrabAlphaSlot (void)
{
	if (r_numalphalist >= MAX_ALPHALIST)
	{
		Con_DPrintf ("Alpha list overflow (this will probably go away after a vis) (unless you're doing something REALLY crazy...)\n");
		return NULL;
	}

	if (!r_alphalist[r_numalphalist])
		r_alphalist[r_numalphalist] = (alphalist_t *) Hunk_Alloc (sizeof (alphalist_t));

	if (!r_alphalist[r_numalphalist]) return NULL;

	r_alphalist[r_numalphalist]->setupfunc = NULL;
	r_alphalist[r_numalphalist]->takedownfunc = NULL;

	r_numalphalist++;
	return r_alphalist[r_numalphalist - 1];
}


void R_CoronasBegin (void);
void R_CoronasEnd (void);
void R_RenderCorona (dlight_t *light);

void R_AddCoronaToAlphaList (dlight_t *l)
{
	alphalist_t *al;

	if (!(al = R_GrabAlphaSlot ())) return;

	al->setupfunc = R_CoronasBegin;
	al->takedownfunc = R_CoronasEnd;
	al->type = ALPHATYPE_CORONA;
	al->dist = R_GetDist (l->origin);
	al->light = l;
}


void R_ParticlesBegin (void);
void R_ParticlesEnd (void);

void R_AddParticlesToAlphaList (particle_type_t *pt)
{
	alphalist_t *al;

	if (!(al = R_GrabAlphaSlot ())) return;

	al->setupfunc = R_ParticlesBegin;
	al->takedownfunc = R_ParticlesEnd;
	al->type = ALPHATYPE_PARTICLE;
	al->dist = R_GetDist (pt->spawnorg);
	al->particles = pt;
}


void R_LiquidBegin (void);
void R_LiquidEnd (void);
void R_SpriteBegin (void);
void R_SpriteEnd (void);

void R_SetupIQMState (void);
void R_TakedownIQMState (void);
void R_DrawIQMModel (entity_t *ent, model_t *mod, iqmdata_t *iqm);

float *R_TransformMidPoint (float *midpoint, glmatrix *m)
{
	static float transformed[3] = {0, 0, 0};

	if (m)
	{
		transformed[0] = midpoint[0] * m->m16[0] + midpoint[1] * m->m16[4] + midpoint[2] * m->m16[8] + m->m16[12];
		transformed[1] = midpoint[0] * m->m16[1] + midpoint[1] * m->m16[5] + midpoint[2] * m->m16[9] + m->m16[13];
		transformed[2] = midpoint[0] * m->m16[2] + midpoint[1] * m->m16[6] + midpoint[2] * m->m16[10] + m->m16[14];
	}
	else
	{
		transformed[0] = midpoint[0];
		transformed[1] = midpoint[1];
		transformed[2] = midpoint[2];
	}

	return transformed;
}


void R_AddLiquidToAlphaList (msurface_t *surf, entity_t *ent, float surfalpha)
{
	alphalist_t *al;

	if (!(al = R_GrabAlphaSlot ())) return;

	al->setupfunc = R_LiquidBegin;
	al->takedownfunc = R_LiquidEnd;

	// to do - translate midpoint by matrix
	al->type = ALPHATYPE_LIQUID;

	if (ent)
		al->dist = R_GetDist (R_TransformMidPoint (surf->midpoint, &ent->matrix));
	else al->dist = R_GetDist (surf->midpoint);

	al->surf = surf;
	al->ent = ent;
	al->surfalpha = surfalpha;
}


void R_ExtraSurfsBegin (void);
void R_ExtraSurfsEnd (void);

void R_AddSurfaceToAlphaList (msurface_t *surf, entity_t *ent, float surfalpha)
{
	alphalist_t *al;

	if (!(al = R_GrabAlphaSlot ())) return;

	al->setupfunc = R_ExtraSurfsBegin;
	al->takedownfunc = R_ExtraSurfsEnd;

	// to do - translate midpoint by matrix
	al->type = ALPHATYPE_SURFACE;
	al->dist = R_GetDist (R_TransformMidPoint (surf->midpoint, &ent->matrix));
	al->surf = surf;
	al->ent = ent;
	al->surfalpha = surfalpha;
}


void R_AddFenceToAlphaList (msurface_t *surf, entity_t *ent, float surfalpha)
{
	alphalist_t *al;

	if (!(al = R_GrabAlphaSlot ())) return;

	al->setupfunc = R_ExtraSurfsBegin;
	al->takedownfunc = R_ExtraSurfsEnd;

	// to do - translate midpoint by matrix
	al->type = ALPHATYPE_FENCE;
	al->dist = R_GetDist (R_TransformMidPoint (surf->midpoint, &ent->matrix));
	al->surf = surf;
	al->ent = ent;
	al->surfalpha = surfalpha;
}


void R_AddEntityToAlphaList (entity_t *ent)
{
	alphalist_t *al;

	if (!(al = R_GrabAlphaSlot ())) return;

	switch (ent->model->type)
	{
	case mod_iqm:
		al->type = ALPHATYPE_IQM;
		al->setupfunc = R_SetupIQMState;
		al->takedownfunc = R_TakedownIQMState;
		// to do...
		break;

	case mod_alias:
		al->type = ALPHATYPE_ALIAS;
		break;

	case mod_brush:
		al->type = ALPHATYPE_BRUSH;
		break;

	case mod_sprite:
		al->type = ALPHATYPE_SPRITE;
		al->setupfunc = R_SpriteBegin;
		al->takedownfunc = R_SpriteEnd;
		break;

	default:
		Host_Error ("R_AddEntityToAlphaList: unknown model type\n");
		return;
	}

	al->dist = R_GetDist (ent->origin);
	al->ent = ent;
}


int R_AlphaSortFunc (const void *a1, const void *a2)
{
	alphalist_t *al1 = *(alphalist_t **) a1;
	alphalist_t *al2 = *(alphalist_t **) a2;

	// back to front ordering
	return (int) (al2->dist - al1->dist);
}


void R_InvalidateSprite (void);
void R_RenderFenceTexture (msurface_t *surf, entity_t *ent, float surfalpha);
void R_DrawAliasBatches (entity_t **ents, int numents, void *meshbuffer);
void R_SetupAliasModel (entity_t *e);
void R_RenderAlphaSurface (msurface_t *surf, entity_t *ent, float surfalpha);
void R_RenderLiquidSurface (msurface_t *surf, entity_t *ent, float surfalpha);


void R_DrawAlphaList (void)
{
	int i;
	alphalist_t *prev;
	alphalist_t *curr;

	if (!r_numalphalist) return;

	if (r_numalphalist > 1)
		qsort (r_alphalist, r_numalphalist, sizeof (alphalist_t *), R_AlphaSortFunc);

	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDepthMask (GL_FALSE);

	glEnable (GL_ALPHA_TEST);
	glAlphaFunc (GL_GREATER, 0.005);

	prev = NULL;

	// invalidate all cached states
	R_InvalidateSprite ();

	for (i = 0; i < r_numalphalist; i++)
	{
		qboolean newtype = false;

		curr = r_alphalist[i];

		if (!prev)
			newtype = true;
		else if (prev->type != curr->type)
		{
			// invalidate all cached states
			R_InvalidateSprite ();

			// callbacks for state takedown (if needed)
			if (prev->takedownfunc)
				prev->takedownfunc ();

			// bring up new (also needed if !prev)
			newtype = true;
		}
		else newtype = false;

		if (newtype && curr->setupfunc)
			curr->setupfunc ();

		// drawing function
		switch (curr->type)
		{
		case ALPHATYPE_IQM:
			R_DrawIQMModel (curr->ent, curr->ent->model, curr->ent->model->iqmdata);
			break;

		case ALPHATYPE_ALIAS:
			GLARB_SetupAliasModel (curr->ent);

			if (curr->ent->visframe == r_framecount)
				GLARB_DrawAliasBatches (&curr->ent, 1);

			break;

		case ALPHATYPE_SPRITE:
			R_DrawSpriteModel (curr->ent);
			break;

		case ALPHATYPE_PARTICLE:
			R_DrawParticlesForType (curr->particles);
			break;

		case ALPHATYPE_FENCE:
			R_RenderFenceTexture (curr->surf, curr->ent, curr->surfalpha);
			break;

		case ALPHATYPE_LIQUID:
			R_RenderLiquidSurface (curr->surf, curr->ent, curr->surfalpha);
			break;

		case ALPHATYPE_SURFACE:
			R_RenderAlphaSurface (curr->surf, curr->ent, curr->surfalpha);
			break;

		case ALPHATYPE_CORONA:
			R_RenderCorona (curr->light);
			break;

		default:
			break;
		}

		prev = curr;
	}

	if (prev->takedownfunc)
		prev->takedownfunc ();

	glAlphaFunc (GL_GREATER, 0.666);
	glDisable (GL_ALPHA_TEST);

	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);

	// not strictly speaking necessary but it does the job well enough
	r_numalphalist = 0;
}


void R_InitNormalMapping (void)
{
}


/*
====================
R_SetClearColor_f -- johnfitz
====================
*/
void R_SetClearColor_f (void)
{
	byte	*rgb;
	int		s;

	s = (int) r_clearcolor.value & 0xFF;
	rgb = (byte *) (d_8to24table + s);
	glClearColor (rgb[0] / 255.0, rgb[1] / 255.0, rgb[2] / 255.0, 0);
}

/*
====================
R_Novis_f -- johnfitz
====================
*/
void R_Novis_f (void)
{
	extern int vis_changed;
	vis_changed = 1;
}

/*
====================
R_OldSkyLeaf_f -- johnfitz
====================
*/
void R_OldSkyLeaf_f (void)
{
	extern int vis_changed;
	vis_changed = 1;
}


/*
===============
R_Envmap_f

Grab six views for environment mapping tests
===============
*/
void R_Envmap_f (void)
{
}

/*
===============
R_Init
===============
*/
void R_WarpRegisterCvars (void);

void R_Init (void)
{
	extern cvar_t r_show_coronas;
	extern cvar_t gl_finish;
	extern cvar_t r_motionblur;
	extern cvar_t r_motionblurradius;
	extern cvar_t r_motionblurfade;
	extern cvar_t r_motionblurupdatetime;
	extern cvar_t r_instancedlight;
	extern cvar_t r_uwfactor;
	extern cvar_t r_uwscale;

	// grrrr.  fucking fragile fitz texture manager strikes again; preferable to just regenerate the textures procedurally here...
	static byte dataw[16][4];
	static byte datag[16][4];
	static byte datab[16][4];
	int i;

	Cmd_AddCommand ("timerefresh", R_TimeRefresh_f);
	Cmd_AddCommand ("envmap", R_Envmap_f);
	Cmd_AddCommand ("pointfile", R_ReadPointFile_f);

	R_WarpRegisterCvars ();

	Cvar_RegisterVariable (&r_uwfactor, NULL);
	Cvar_RegisterVariable (&r_uwscale, NULL);

	Cvar_RegisterVariable (&r_colorfilter, NULL);
	Cvar_RegisterVariable (&r_motionblur, NULL);
	Cvar_RegisterVariable (&r_motionblurradius, NULL);
	Cvar_RegisterVariable (&r_motionblurfade, NULL);
	Cvar_RegisterVariable (&r_motionblurupdatetime, NULL);
	Cvar_RegisterVariable (&r_show_coronas, NULL);
	Cvar_RegisterVariable (&r_showtris, NULL);

	Cvar_RegisterVariable (&r_norefresh, NULL);
	Cvar_RegisterVariable (&r_lightmap, NULL);
	Cvar_RegisterVariable (&r_fullbright, NULL);
	Cvar_RegisterVariable (&r_drawentities, NULL);
	Cvar_RegisterVariable (&r_drawviewmodel, NULL);
	Cvar_RegisterVariable (&r_shadows, NULL);
	Cvar_RegisterVariable (&r_wateralpha, NULL);
	Cvar_RegisterVariable (&r_dynamic, NULL);
	Cvar_RegisterVariable (&r_novis, R_Novis_f);
	Cvar_RegisterVariable (&r_speeds, NULL);

	Cvar_RegisterVariable (&gl_finish, NULL);
	Cvar_RegisterVariable (&gl_clear, NULL);
	Cvar_RegisterVariable (&gl_cull, NULL);
	Cvar_RegisterVariable (&gl_smoothmodels, NULL);
	Cvar_RegisterVariable (&gl_affinemodels, NULL);
	Cvar_RegisterVariable (&gl_polyblend, NULL);
	Cvar_RegisterVariable (&gl_playermip, NULL);
	Cvar_RegisterVariable (&gl_nocolors, NULL);

	// johnfitz -- new cvars
	Cvar_RegisterVariable (&r_clearcolor, R_SetClearColor_f);
	Cvar_RegisterVariable (&r_waterwarp, NULL);
	Cvar_RegisterVariable (&r_drawflat, NULL);
	Cvar_RegisterVariable (&r_flatlightstyles, NULL);
	Cvar_RegisterVariable (&r_oldskyleaf, R_OldSkyLeaf_f);
	Cvar_RegisterVariable (&r_drawworld, NULL);
	Cvar_RegisterVariable (&r_showbboxes, NULL);
	Cvar_RegisterVariable (&gl_fullbrights, NULL);
	Cvar_RegisterVariable (&r_lerpmodels, NULL);
	Cvar_RegisterVariable (&r_lerpmove, NULL);
	Cvar_RegisterVariable (&r_nolerp_list, R_NoLerpList_f);
	// johnfitz

	Cvar_RegisterVariable (&gl_src_blend, NULL);
	Cvar_RegisterVariable (&gl_dst_blend, NULL);

	Cvar_RegisterVariable (&r_lockpvs, NULL);
	Cvar_RegisterVariable (&r_lockfrustum, NULL);
	Cvar_RegisterVariable (&r_normalmapping, NULL);

	Cvar_RegisterVariable (&r_instancedlight, NULL);

	R_InitParticles ();
	R_SetClearColor_f (); // johnfitz

	Sky_Init (); // johnfitz
	Fog_Init (); // johnfitz

	// these textures must have the same value in alpha as they have in RGB for lighting to work properly
	for (i = 0; i < 16; i++)
	{
		dataw[i][0] = 128;
		dataw[i][1] = 128;
		dataw[i][2] = 128;
		dataw[i][3] = 128;
	}

	r_whitetexture = TM_LoadImage (NULL, "white", 4, 4, SRC_RGBA, dataw[0], "", (src_offset_t) dataw, TEXPREF_NONE);

	for (i = 0; i < 16; i++)
	{
		datag[i][0] = 128;
		datag[i][1] = 128;
		datag[i][2] = 128;
		datag[i][3] = 128;
	}

	r_greytexture = TM_LoadImage (NULL, "grey", 4, 4, SRC_RGBA, datag[0], "", (src_offset_t) datag, TEXPREF_NONE);

	for (i = 0; i < 16; i++)
	{
		datab[i][0] = 0;
		datab[i][1] = 0;
		datab[i][2] = 0;
		datab[i][3] = 0;
	}

	r_blacktexture = TM_LoadImage (NULL, "black", 4, 4, SRC_RGBA, datab[0], "", (src_offset_t) datab, TEXPREF_NONE);
}


/*
===============
R_NoLerpList_f -- johnfitz -- called when r_nolerp_list cvar changes
===============
*/
void R_NoLerpList_f (void)
{
	int i;

	for (i = 0; i < MAX_MODELS; i++)
		Mod_SetExtraFlags (cl.model_precache[i]);
}

/*
===============
R_TranslatePlayerSkin -- johnfitz -- rewritten.  also, only handles new colors, not new skins
===============
*/
void R_TranslatePlayerSkin (int playernum)
{
	int			top, bottom;

	top = (cl.scores[playernum].colors & 0xf0) >> 4;
	bottom = cl.scores[playernum].colors & 15;

	// FIXME: if gl_nocolors is on, then turned off, the textures may be out of sync with the scoreboard colors.
	if (!gl_nocolors.value)
		if (playertextures[playernum])
			TM_ReloadImage (playertextures[playernum], top, bottom);
}

/*
===============
R_TranslateNewPlayerSkin -- johnfitz -- split off of TranslatePlayerSkin -- this is called when
the skin or model actually changes, instead of just new colors
added bug fix from bengt jardup
===============
*/
void R_TranslateNewPlayerSkin (int playernum)
{
	char		name[64];
	byte		*pixels;
	aliashdr_t	*paliashdr;
	int			skinnum;
	entity_t	*e;

	// get correct texture pixels
	e = cl_entities[1+playernum];

	if (!e->model || e->model->type != mod_alias)
		return;

	paliashdr = (aliashdr_t *) Mod_Extradata (e->model);

	skinnum = e->skinnum;

	// TODO: move these tests to the place where skinnum gets received from the server
	if (skinnum < 0 || skinnum >= paliashdr->numskins)
	{
		Con_DPrintf ("(%d): Invalid player skin #%d\n", playernum, skinnum);
		skinnum = 0;
	}

	pixels = (byte *) paliashdr + paliashdr->texels[skinnum]; // This is not a persistent place!

	// upload new image
	sprintf (name, "player_%i", playernum);
	playertextures[playernum] = TM_LoadImage (e->model, name, paliashdr->skinwidth, paliashdr->skinheight,
								SRC_INDEXED, pixels, paliashdr->gltextures[skinnum][0]->source_file, paliashdr->gltextures[skinnum][0]->source_offset, TEXPREF_PAD | TEXPREF_OVERWRITE);

	// now recolor it
	R_TranslatePlayerSkin (playernum);
}

/*
===============
R_NewGame -- johnfitz -- handle a game switch
===============
*/
void R_NewGame (void)
{
	int i;

	// clear playertexture pointers (the textures themselves were freed by texmgr_newgame)
	for (i = 0; i < MAX_SCOREBOARD; i++)
		playertextures[i] = NULL;
}

/*
===============
R_NewMap
===============
*/
// mh 2010-07-16 - clear down modelsurfs from the previous map
void R_InitWorldCaches (void);
void WaterAlpha_ParseWorldspawn (void);
extern char mod_worldname[];
void RBrush_LoadVertexBuffers (void);
void GLMesh_LoadVertexBuffers (void);
void GLIQM_LoadVertexBuffers (void);

extern qboolean r_rebuildworld;

void R_NewMap (void)
{
	int		i;
	int start, end;

	start = SDL_GetTicks ();

	// mh 2010-07-16 - clear down stuff from the previous map
	R_InitWorldCaches ();
	R_AlphaListBeginMap ();

	// clear out efrags in case the level hasn't been reloaded
	// FIXME: is this one short?
	for (i = 0; i < cl.worldmodel->numleafs; i++)
		cl.worldmodel->leafs[i].efrags = NULL;

	r_viewleaf = NULL;
	R_ClearParticles ();

	GL_BuildLightmaps ();

	// prevent all leafs, nodes and surfs from being added to the view in the first frame
	//r_framecount = 0; // johnfitz -- paranoid?
	//r_visframecount = 0; // johnfitz -- paranoid?

	Sky_NewMap (); // johnfitz -- skybox in worldspawn
	Fog_NewMap (); // johnfitz -- global fog in worldspawn
	// WaterAlpha_ParseWorldspawn ();

	// switch off all view blending
	v_blend[0] = v_blend[1] = v_blend[2] = v_blend[3] = 0;

	// used to track items registered for this map
	// go to the next registration sequence
	r_registrationsequence++;

	// for the next map
	mod_worldname[0] = 0;

	// force surfaces to recache on map entry
	r_rebuildworld = true;

	// reload all of our vertex buffers
	GL_BeginBuffers ();
	RBrush_LoadVertexBuffers ();
	GLMesh_LoadVertexBuffers ();
	GLIQM_LoadVertexBuffers ();

	// switch off damage blur
	cl.damageblur = -1;

	// clear all color shifts
	memset (cl.cshifts, 0, sizeof (cshift_t) * NUM_CSHIFTS);
	memset (cl.prev_cshifts, 0, sizeof (cshift_t) * NUM_CSHIFTS);

	end = SDL_GetTicks ();

	Con_DPrintf ("R_NewMap in %i ms\n", (end - start));
}


/*
====================
R_TimeRefresh_f

For program optimization
====================
*/
void R_TimeRefresh_f (void)
{
}

void D_FlushCaches (void)
{
}


