/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// sv_edict.c -- entity dictionary

#include "quakedef.h"

static void PR_AllocStringSlots (void);

prvm_t cl_progs;	// server-side vm
prvm_t sv_progs;	// server-side vm
prvm_t *pr_progs;	// currently active vm

int type_size[8] =
{
	1,					// ev_void
	1,	// sizeof(string_t) / 4		// ev_string
	1,					// ev_float
	3,					// ev_vector
	1,					// ev_entity
	1,					// ev_field
	1,	// sizeof(func_t) / 4		// ev_function
	1	// sizeof(void *) / 4		// ev_pointer
};

ddef_t *ED_FieldAtOfs (int ofs);
qboolean	ED_ParseEpair (void *base, ddef_t *key, char *s);

cvar_t	nomonsters = {"nomonsters", "0"};
cvar_t	gamecfg = {"gamecfg", "0"};
cvar_t	scratch1 = {"scratch1", "0"};
cvar_t	scratch2 = {"scratch2", "0"};
cvar_t	scratch3 = {"scratch3", "0"};
cvar_t	scratch4 = {"scratch4", "0"};
cvar_t	savedgamecfg = {"savedgamecfg", "0", true};
cvar_t	saved1 = {"saved1", "0", true};
cvar_t	saved2 = {"saved2", "0", true};
cvar_t	saved3 = {"saved3", "0", true};
cvar_t	saved4 = {"saved4", "0", true};

// 2001-09-14 Enhanced BuiltIn Function System (EBFS) by Maddes  start
cvar_t	pr_builtin_find = {"pr_builtin_find", "0", false, false};
cvar_t	pr_builtin_remap = {"pr_builtin_remap", "0", false, false};
// 2001-09-14 Enhanced BuiltIn Function System (EBFS) by Maddes  end

cvar_t	pr_checkextension = {"pr_checkextension", "0", false, false};	// 2001-10-20 Extension System by LordHavoc (DP compatibility)


void ED_Memset0 (edict_t *e)
{
	// any memset 0 on an edict_t should use this instead of memsetting directly so that field values can be fixed up properly
	memset (&e->v, 0, pr_progs->progs->entityfields * 4);

	// fix instanity with fitz alpha
	if (sv.protocol == PROTOCOL_RMQ && (sv.protocolflags & PRFL_ALPHASANITY))
	{
		if (eval_alpha)
			((eval_t *) ((char *) &e->v + eval_alpha))->_float = 255.0f;
		e->alpha = 255;
	}
	else
	{
		e->alpha = ENTALPHA_DEFAULT;
	}

	e->scale = 16;
}


/*
=================
ED_ClearEdict

Sets everything to NULL
=================
*/
void ED_ClearEdict (edict_t *e)
{
	ED_Memset0 (e);
	e->free = false;
}

/*
=================
ED_Alloc

Either finds a free edict, or allocates a new one.
Try to avoid reusing an entity that was recently freed, because it
can cause the client to think the entity morphed into something else
instead of being removed and recreated, which can cause interpolated
angles and bad trails.
=================
*/
edict_t *ED_Alloc (void)
{
	int			i;
	edict_t		*e;

	// start on the edict after the clients
	for (i = pr_progs->first_edict; i < pr_progs->num_edicts; i++)
	{
		e = GetEdictForNumber (i);

		// the first couple seconds of server time can involve a lot of
		// freeing and allocating, so relax the replacement policy
		if (e->free && (e->freetime < 2.0f || pr_progs->time - e->freetime > 0.5f))
		{
			ED_ClearEdict (e);
			return e;
		}
	}

	if (i >= MAX_EDICTS)
	{
		// if we hit the absolute upper limit just pick the one with the lowest free time
		float bestfree = pr_progs->time;
		int bestedict = -1;

		for (i = pr_progs->first_edict; i < pr_progs->num_edicts; i++)
		{
			e = GetEdictForNumber (i);

			if (e->free && e->freetime < bestfree)
			{
				bestedict = i;
				bestfree = e->freetime;
			}
		}

		if (bestedict > -1)
		{
			e = GetEdictForNumber (bestedict);
			ED_ClearEdict (e);

			return e;
		}

		// no free edicts at all!!!
		Host_Error ("ED_Alloc: edict count at protocol maximum!");
	}

	// alloc 32 more edicts
	if (i >= pr_progs->max_edicts) SV_AllocEdicts (32);

	pr_progs->num_edicts++;
	e = GetEdictForNumber (i);
	ED_ClearEdict (e);

	return e;
}

/*
=================
ED_Free

Marks the edict as free
FIXME: walk all entities and NULL out references to this entity
=================
*/
void ED_Free (edict_t *ed)
{
	// fixme - what happens if we call ED_Free from csprogs???
	SV_UnlinkEdict (ed);		// unlink from world bsp

	ed->free = true;
	ed->v.model = 0;
	ed->v.takedamage = 0;
	ed->v.modelindex = 0;
	ed->v.colormap = 0;
	ed->v.skin = 0;
	ed->v.frame = 0;
	VectorCopy (vec3_origin, ed->v.origin);
	VectorCopy (vec3_origin, ed->v.angles);
	ed->v.nextthink = -1;
	ed->v.solid = 0;
	ed->alpha = ENTALPHA_DEFAULT; // johnfitz -- reset alpha for next entity

	ed->freetime = pr_progs->time;
}

// ===========================================================================

/*
============
ED_GlobalAtOfs
============
*/
ddef_t *ED_GlobalAtOfs (int ofs)
{
	ddef_t		*def;
	int			i;

	for (i = 0; i < pr_progs->progs->numglobaldefs; i++)
	{
		def = &pr_progs->globaldefs[i];

		if (def->ofs == ofs)
			return def;
	}

	return NULL;
}

/*
============
ED_FieldAtOfs
============
*/
ddef_t *ED_FieldAtOfs (int ofs)
{
	ddef_t		*def;
	int			i;

	for (i = 0; i < pr_progs->progs->numfielddefs; i++)
	{
		def = &pr_progs->fielddefs[i];

		if (def->ofs == ofs)
			return def;
	}

	return NULL;
}

/*
============
ED_FindField
============
*/
ddef_t *ED_FindField (char *name)
{
	ddef_t		*def;
	int			i;

	for (i = 0; i < pr_progs->progs->numfielddefs; i++)
	{
		def = &pr_progs->fielddefs[i];

		if (!strcmp (PR_GetString (def->s_name), name))
			return def;
	}

	return NULL;
}


// fast eval lookups for stuff we get regularly at runtime
// these need to move to prvm_t
int eval_nodrawtoclient; // Team Xlink DP_SV_NODRAWTOCLIENT
int eval_drawonlytoclient; // Team Xlink DP_SV_DRAWONLYTOCLIENT
int eval_clientcamera; // DP_SV_CLIENTCAMERA
int eval_alpha;
int eval_scale;

int FindFieldOffset(char *field)
{
   ddef_t *d;
   d = ED_FindField (field);
   if (!d)
      return 0;
   return d->ofs*4;
}

void FindEdictFieldOffsets ()
{
	eval_nodrawtoclient = FindFieldOffset ("nodrawtoclient");
	eval_drawonlytoclient = FindFieldOffset ("drawonlytoclient");
	eval_clientcamera = FindFieldOffset ("clientcamera");	// DP_SV_CLIENTCAMERA

	eval_alpha = FindFieldOffset ("alpha");
	eval_scale = FindFieldOffset ("scale");
};


/*
============
ED_FindGlobal
============
*/
ddef_t *ED_FindGlobal (char *name)
{
	ddef_t		*def;
	int			i;

	for (i = 0; i < pr_progs->progs->numglobaldefs; i++)
	{
		def = &pr_progs->globaldefs[i];

		if (!strcmp (PR_GetString (def->s_name), name))
			return def;
	}

	return NULL;
}


/*
============
ED_FindFunction
============
*/
dfunction_t *ED_FindFunction (char *name)
{
	dfunction_t		*func;
	int				i;

	for (i = 0; i < pr_progs->progs->numfunctions; i++)
	{
		func = &pr_progs->functions[i];

		if (!strcmp (PR_GetString (func->s_name), name))
			return func;
	}

	return NULL;
}

/*
============
GetEdictFieldValue
============
*/
eval_t *GetEdictFieldValue (edict_t *ed, char *field)
{
	ddef_t			*def = NULL;
	int				i;
	static int		rep = 0;

	for (i = 0; i < GEFV_CACHESIZE; i++)
	{
		if (!strcmp (field, pr_progs->GEFV_Cache[i].field))
		{
			def = pr_progs->GEFV_Cache[i].pcache;
			goto Done;
		}
	}

	def = ED_FindField (field);

	if (strlen (field) < MAX_FIELD_LEN)
	{
		pr_progs->GEFV_Cache[rep].pcache = def;
		strcpy (pr_progs->GEFV_Cache[rep].field, field);
		rep ^= 1;
	}

Done:;
	return def ? ((eval_t *) ((char *) &ed->v + def->ofs * 4)) : NULL;
}


/*
============
PR_ValueString

Returns a string describing *data in a type specific manner
=============
*/
char *PR_ValueString (etype_t type, eval_t *val)
{
	static char	line[256];
	ddef_t		*def;
	dfunction_t	*f;

	type &= ~DEF_SAVEGLOBAL;

	switch (type)
	{
	case ev_string:
		sprintf (line, "%s", PR_GetString (val->string));
		break;
	case ev_entity:
		sprintf (line, "entity %i", GetNumberForEdict (PROG_TO_EDICT (val->edict)));
		break;
	case ev_function:
		f = pr_progs->functions + val->function;
		sprintf (line, "%s ()", PR_GetString (f->s_name));
		break;
	case ev_field:
		def = ED_FieldAtOfs (val->_int);
		sprintf (line, ".%s", PR_GetString (def->s_name));
		break;
	case ev_void:
		sprintf (line, "void");
		break;
	case ev_float:
		sprintf (line, "%5.1f", val->_float);
		break;
	case ev_vector:
		sprintf (line, "'%5.1f %5.1f %5.1f'", val->vector[0], val->vector[1], val->vector[2]);
		break;
	case ev_pointer:
		sprintf (line, "pointer");
		break;
	default:
		sprintf (line, "bad type %i", type);
		break;
	}

	return line;
}

/*
============
PR_UglyValueString

Returns a string describing *data in a type specific manner
Easier to parse than PR_ValueString
=============
*/
char *PR_UglyValueString (etype_t type, eval_t *val)
{
	static char	line[256];
	ddef_t		*def;
	dfunction_t	*f;

	type &= ~DEF_SAVEGLOBAL;

	switch (type)
	{
	case ev_string:
		sprintf (line, "%s", PR_GetString (val->string));
		break;
	case ev_entity:
		sprintf (line, "%i", GetNumberForEdict (PROG_TO_EDICT (val->edict)));
		break;
	case ev_function:
		f = pr_progs->functions + val->function;
		sprintf (line, "%s", PR_GetString (f->s_name));
		break;
	case ev_field:
		def = ED_FieldAtOfs (val->_int);
		sprintf (line, "%s", PR_GetString (def->s_name));
		break;
	case ev_void:
		sprintf (line, "void");
		break;
	case ev_float:
		sprintf (line, "%f", val->_float);
		break;
	case ev_vector:
		sprintf (line, "%f %f %f", val->vector[0], val->vector[1], val->vector[2]);
		break;
	default:
		sprintf (line, "bad type %i", type);
		break;
	}

	return line;
}

/*
============
PR_GlobalString

Returns a string with a description and the contents of a global,
padded to 20 field width
============
*/
char *PR_GlobalString (int ofs)
{
	char	*s;
	int		i;
	ddef_t	*def;
	void	*val;
	static char	line[128];

	val = (void *) &pr_progs->globals[ofs];
	def = ED_GlobalAtOfs (ofs);

	if (!def)
	{
		// silly hack to get rid of trigraphs warning
		sprintf (line, "%i(?%c?)", ofs, 63);
	}
	else
	{
		s = PR_ValueString (def->type, (eval_t *) val);
		sprintf (line, "%i(%s)%s", ofs, PR_GetString (def->s_name), s);
	}

	i = strlen (line);

	for (; i < 20; i++)
		strcat (line, " ");

	strcat (line, " ");

	return line;
}

char *PR_GlobalStringNoContents (int ofs)
{
	int		i;
	ddef_t	*def;
	static char	line[128];

	def = ED_GlobalAtOfs (ofs);

	if (!def)
	{
		// more trigraph warnings here
		sprintf (line, "%i(?%c?)", ofs, 63);
	}
	else
		sprintf (line, "%i(%s)", ofs, PR_GetString (def->s_name));

	i = strlen (line);

	for (; i < 20; i++)
		strcat (line, " ");

	strcat (line, " ");

	return line;
}


/*
=============
ED_Print

For debugging
=============
*/
void ED_Print (edict_t *ed)
{
	ddef_t	*d;
	int		*v;
	int		i, j, l;
	char	*name;
	int		type;

	if (ed->free)
	{
		Con_Printf ("FREE\n");
		return;
	}

	Con_SafePrintf ("\nEDICT %i:\n", GetNumberForEdict (ed)); // johnfitz -- was Con_Printf

	for (i = 1; i < pr_progs->progs->numfielddefs; i++)
	{
		d = &pr_progs->fielddefs[i];
		name = PR_GetString (d->s_name);
		l = strlen (name);

		if (l > 1 && name[l - 2] == '_')
			continue;	// skip _x, _y, _z vars

		v = (int *) ((char *) &ed->v + d->ofs * 4);

		// if the value is still all 0, skip the field
		type = d->type & ~DEF_SAVEGLOBAL;

		for (j = 0; j < type_size[type]; j++)
			if (v[j])
				break;

		if (j == type_size[type])
			continue;

		Con_SafePrintf ("%s", name); // johnfitz -- was Con_Printf

		while (l++ < 15)
			Con_SafePrintf (" "); // johnfitz -- was Con_Printf

		Con_SafePrintf ("%s\n", PR_ValueString (d->type, (eval_t *) v)); // johnfitz -- was Con_Printf
	}
}

/*
=============
ED_Write

For savegames
=============
*/
void ED_Write (FILE *f, edict_t *ed)
{
	ddef_t	*d;
	int		*v;
	int		i, j;
	char	*name;
	int		type;

	fprintf (f, "{\n");

	if (ed->free)
	{
		fprintf (f, "}\n");
		return;
	}

	for (i = 1; i < pr_progs->progs->numfielddefs; i++)
	{
		d = &pr_progs->fielddefs[i];
		name = PR_GetString (d->s_name);
		j = strlen (name);

		if (j > 1 && name[j - 2] == '_')
			continue;	// skip _x, _y, _z vars

		v = (int *) ((char *) &ed->v + d->ofs * 4);

		// if the value is still all 0, skip the field
		type = d->type & ~DEF_SAVEGLOBAL;

		for (j = 0; j < type_size[type]; j++)
			if (v[j])
				break;

		if (j == type_size[type])
			continue;

		fprintf (f, "\"%s\" ", name);
		fprintf (f, "\"%s\"\n", PR_UglyValueString (d->type, (eval_t *) v));
	}

	// johnfitz -- save entity alpha manually when progs.dat doesn't know about alpha
	// ????? is this a little bit silly ?????
	if (!eval_alpha && ed->alpha != ENTALPHA_DEFAULT)
		fprintf (f, "\"alpha\" \"%f\"\n", ENTALPHA_TOSAVE (ed->alpha));
	// johnfitz

	fprintf (f, "}\n");
}

void ED_PrintNum (int ent)
{
	ED_Print (GetEdictForNumber (ent));
}

/*
=============
ED_PrintEdicts

For debugging, prints all the entities in the current server
=============
*/
void ED_PrintEdicts (void)
{
	int		i;

	Con_Printf ("%i entities\n", pr_progs->num_edicts);

	for (i = 0; i < pr_progs->num_edicts; i++)
		ED_PrintNum (i);
}

/*
=============
ED_PrintEdict_f

For debugging, prints a single edicy
=============
*/
void ED_PrintEdict_f (void)
{
	int		i;

	i = Q_atoi (Cmd_Argv (1));

	if (i >= pr_progs->num_edicts)
	{
		Con_Printf ("Bad edict number\n");
		return;
	}

	ED_PrintNum (i);
}

/*
=============
ED_Count

For debugging
=============
*/
void ED_Count (void)
{
	int		i;
	edict_t	*ent;
	int		active, models, solid, step;

	active = models = solid = step = 0;

	for (i = 0; i < pr_progs->num_edicts; i++)
	{
		ent = GetEdictForNumber (i);

		if (ent->free)
			continue;

		active++;

		if (ent->v.solid)
			solid++;

		if (ent->v.model)
			models++;

		if (ent->v.movetype == MOVETYPE_STEP)
			step++;
	}

	Con_Printf ("num_edicts:%3i\n", pr_progs->num_edicts);
	Con_Printf ("active    :%3i\n", active);
	Con_Printf ("view      :%3i\n", models);
	Con_Printf ("touch     :%3i\n", solid);
	Con_Printf ("step      :%3i\n", step);

}

/*
==============================================================================

					ARCHIVING GLOBALS

FIXME: need to tag constants, doesn't really work
==============================================================================
*/

/*
=============
ED_WriteGlobals
=============
*/
void ED_WriteGlobals (FILE *f)
{
	ddef_t		*def;
	int			i;
	char		*name;
	int			type;

	fprintf (f, "{\n");

	for (i = 0; i < pr_progs->progs->numglobaldefs; i++)
	{
		def = &pr_progs->globaldefs[i];
		type = def->type;

		if (!(def->type & DEF_SAVEGLOBAL))
			continue;

		type &= ~DEF_SAVEGLOBAL;

		if (type != ev_string
		&& type != ev_float
		&& type != ev_entity)
			continue;

		name = PR_GetString (def->s_name);
		fprintf (f, "\"%s\" ", name);
		fprintf (f, "\"%s\"\n", PR_UglyValueString (type, (eval_t *) &pr_progs->globals[def->ofs]));
	}

	fprintf (f, "}\n");
}

/*
=============
ED_ParseGlobals
=============
*/
void ED_ParseGlobals (char *data)
{
	char	keyname[64];
	ddef_t	*key;

	while (1)
	{
		// parse key
		data = COM_Parse (data);

		if (com_token[0] == '}')
			break;

		if (!data)
			Sys_Error ("ED_ParseEntity: EOF without closing brace");

		strcpy (keyname, com_token);

		// parse value
		data = COM_Parse (data);

		if (!data)
			Sys_Error ("ED_ParseEntity: EOF without closing brace");

		if (com_token[0] == '}')
			Sys_Error ("ED_ParseEntity: closing brace without data");

		key = ED_FindGlobal (keyname);

		if (!key)
		{
			Con_Printf ("'%s' is not a global\n", keyname);
			continue;
		}

		if (!ED_ParseEpair ((void *) pr_progs->globals, key, com_token))
			Host_Error ("ED_ParseGlobals: parse error");
	}
}

// ============================================================================


int PR_AllocString (int size, char **ptr)
{
	int		i;

	if (!size)
		return 0;

	for (i = 0; i < pr_progs->numknownstrings; i++)
	{
		if (!pr_progs->knownstrings[i])
			break;
	}

	// 	if (i >= pr_progs->numknownstrings)
	// 	{
	if (i >= pr_progs->maxknownstrings)
		PR_AllocStringSlots ();

	pr_progs->numknownstrings++;
	// 	}
	pr_progs->knownstrings[i] = (char *) Hunk_AllocName (size, "string");

	if (ptr)
		*ptr = pr_progs->knownstrings[i];

	return -1 - i;
}


/*
=============
ED_NewString
=============
*/
static string_t ED_NewString (const char *string)
{
	char	*new_p;
	int		i, l;
	string_t	num;

	l = strlen (string) + 1;
	num = PR_AllocString (l, &new_p);

	for (i = 0; i < l; i++)
	{
		if (string[i] == '\\' && i < l - 1)
		{
			i++;

			if (string[i] == 'n')
				*new_p++ = '\n';
			else
				*new_p++ = '\\';
		}
		else
			*new_p++ = string[i];
	}

	return num;
}


/*
=============
ED_ParseEval

Can parse either fields or globals
returns false if error
=============
*/
qboolean	ED_ParseEpair (void *base, ddef_t *key, char *s)
{
	int		i;
	char	string[128];
	ddef_t	*def;
	char	*v, *w;
	void	*d;
	dfunction_t	*func;

	d = (void *) ((int *) base + key->ofs);

	switch (key->type & ~DEF_SAVEGLOBAL)
	{
	case ev_string:
		*(string_t *) d = ED_NewString (s);
		break;

	case ev_float:
		*(float *) d = atof (s);
		break;

	case ev_vector:
		strcpy (string, s);
		v = string;
		w = string;

		for (i = 0; i < 3; i++)
		{
			while (*v && *v != ' ')
				v++;

			*v = 0;
			((float *) d)[i] = atof (w);
			w = v = v + 1;
		}

		break;

	case ev_entity:
		*(int *) d = EDICT_TO_PROG (GetEdictForNumber (atoi (s)));
		break;

	case ev_field:
		def = ED_FindField (s);

		if (!def)
		{
			// johnfitz -- HACK -- suppress error becuase fog/sky fields might not be mentioned in defs.qc
			if (strncmp (s, "sky", 3) && strcmp (s, "fog"))
				Con_DPrintf ("Can't find field %s\n", s);

			return false;
		}

		*(int *) d = G_INT (def->ofs);
		break;

	case ev_function:
		func = ED_FindFunction (s);

		if (!func)
		{
			Con_Printf ("Can't find function %s\n", s);
			return false;
		}

		*(func_t *) d = func - pr_progs->functions;
		break;

	default:
		break;
	}

	return true;
}

/*
====================
ED_ParseEdict

Parses an edict out of the given string, returning the new position
ed should be a properly initialized empty edict.
Used for initial level load and for savegames.
====================
*/
char *ED_ParseEdict (char *data, edict_t *ent)
{
	ddef_t		*key;
	qboolean	anglehack;
	qboolean	init;
	char		keyname[256];
	int			n;

	init = false;

	// clear it
	if (ent != pr_progs->edicts[0])	// hack
		ED_Memset0 (ent);

	// default scale
	ent->scale = 16;

	// go through all the dictionary pairs
	while (1)
	{
		// parse key
		data = COM_Parse (data);

		if (com_token[0] == '}')
			break;

		if (!data)
			Sys_Error ("ED_ParseEntity: EOF without closing brace");

		// anglehack is to allow QuakeEd to write single scalar angles
		// and allow them to be turned into vectors. (FIXME...)
		if (!strcmp (com_token, "angle"))
		{
			strcpy (com_token, "angles");
			anglehack = true;
		}
		else anglehack = false;

		// FIXME: change light to _light to get rid of this hack
		if (!strcmp (com_token, "light"))
			strcpy (com_token, "light_lev");	// hack for single light def

		strcpy (keyname, com_token);

		// another hack to fix keynames with trailing spaces
		n = strlen (keyname);

		while (n && keyname[n-1] == ' ')
		{
			keyname[n-1] = 0;
			n--;
		}

		// parse value
		data = COM_Parse (data);

		if (!data) Sys_Error ("ED_ParseEntity: EOF without closing brace");
		if (com_token[0] == '}') Sys_Error ("ED_ParseEntity: closing brace without data");

		init = true;

		// keynames with a leading underscore are used for utility comments,
		// and are immediately discarded by quake
		if (keyname[0] == '_')
			continue;

		// johnfitz -- hack to support .alpha even when progs.dat doesn't know about it
		if (!strcmp (keyname, "alpha")) ent->alpha = ENTALPHA_ENCODE (atof (com_token));
		// johnfitz

		key = ED_FindField (keyname);

		if (!key)
		{
			// johnfitz -- HACK -- suppress error becuase fog/sky/alpha fields might not be mentioned in defs.qc
			if (strncmp (keyname, "sky", 3) && strcmp (keyname, "fog") && strcmp (keyname, "alpha") && strcmp (keyname, "scale"))
				Con_DPrintf ("\"%s\" is not a field\n", keyname); // johnfitz -- was Con_Printf

			continue;
		}

		if (anglehack)
		{
			char	temp[32];
			strcpy (temp, com_token);
			sprintf (com_token, "0 %s 0", temp);
		}

		if (!ED_ParseEpair ((void *) &ent->v, key, com_token))
			Host_Error ("ED_ParseEdict: parse error");
	}

	if (!init)
		ent->free = true;

	return data;
}


/*
================
ED_LoadFromFile

The entities are directly placed in the array, rather than allocated with
ED_Alloc, because otherwise an error loading the map would have entity
number references out of order.

Creates a server's entity / program execution context by
parsing textual entity definitions out of an ent file.

Used for both fresh maps and savegame loads.  A fresh map would also need
to call ED_CallSpawnFunctions () to let the objects initialize themselves.
================
*/
void ED_LoadFromFile (char *data)
{
	edict_t		*ent = NULL;
	int		inhibit = 0;
	dfunction_t	*func;

	pr_progs->global_struct->time = pr_progs->time;

	// parse ents
	while (1)
	{
		// parse the opening brace
		data = COM_Parse (data);

		if (!data)
			break;

		if (com_token[0] != '{')
			Sys_Error ("ED_LoadFromFile: found %s when expecting {", com_token);

		if (!ent)
			ent = GetEdictForNumber (0);
		else ent = ED_Alloc ();

		data = ED_ParseEdict (data, ent);

		// remove things from different skill levels or deathmatch
		if (deathmatch.value)
		{
			if (((int) ent->v.spawnflags & SPAWNFLAG_NOT_DEATHMATCH))
			{
				ED_Free (ent);
				inhibit++;
				continue;
			}
		}
		else if ((current_skill == 0 && ((int) ent->v.spawnflags & SPAWNFLAG_NOT_EASY))
		|| (current_skill == 1 && ((int) ent->v.spawnflags & SPAWNFLAG_NOT_MEDIUM))
		|| (current_skill >= 2 && ((int) ent->v.spawnflags & SPAWNFLAG_NOT_HARD)))
		{
			ED_Free (ent);
			inhibit++;
			continue;
		}

		//
		// immediately call spawn function
		//
		if (!ent->v.classname)
		{
			Con_SafePrintf ("No classname for:\n"); // johnfitz -- was Con_Printf
			ED_Print (ent);
			ED_Free (ent);
			continue;
		}

		// look for the spawn function
		func = ED_FindFunction (PR_GetString (ent->v.classname));

		if (!func)
		{
			Con_SafePrintf ("No spawn function for:\n"); // johnfitz -- was Con_Printf
			ED_Print (ent);
			ED_Free (ent);
			continue;
		}

		pr_progs->global_struct->self = EDICT_TO_PROG (ent);
		PR_ExecuteProgram (&sv_progs, func - pr_progs->functions);
	}

	Con_DPrintf ("%i entities inhibited\n", inhibit);
}


/*
===============
PR_LoadProgs
===============
*/
qboolean PR_LoadProgs (prvm_t *progs, char *filename, qboolean mandatory, int crc)
{
	int i;

// 2001-09-14 Enhanced BuiltIn Function System (EBFS) by Maddes/Firestorm  start
	int 	j;
	int		funcno;
	char	*funcname;
// 2001-09-14 Enhanced BuiltIn Function System (EBFS) by Maddes/Firestorm  end

	int start, end;

	// for saving out
	ebfs_builtin_t *ebfs_builtins;
	int ebfs_numbuiltins;

	// ensure that the currently active progs is set in case anything here needs to call functions that depend on it
	PR_SetActiveProgs (progs);

	// these are done before the memset 0 so that we don't leak
	// let's not abuse zone memory
	if (pr_progs->knownstrings)
		free (pr_progs->knownstrings);

	// save out the builtins table so that memset 0 doesn't hose it
	ebfs_builtins = pr_progs->ebfs_builtins;
	ebfs_numbuiltins = pr_progs->ebfs_numbuiltins;

	// clear out progs pointers so that nothing from the previous load is left hanging around
	memset (pr_progs, 0, sizeof (prvm_t));

	// now restore the builtins
	pr_progs->ebfs_builtins = ebfs_builtins;
	pr_progs->ebfs_numbuiltins = ebfs_numbuiltins;

	start = SDL_GetTicks ();

	// flush the non-C variable lookup cache
	for (i = 0; i < GEFV_CACHESIZE; i++)
	{
		pr_progs->GEFV_Cache[i].field[0] = 0;
		pr_progs->GEFV_Cache[i].pcache = NULL;
	}

	CRC_Init (&pr_progs->crc);

	pr_progs->progs = (dprograms_t *) COM_LoadHunkFile (filename);

	if (!pr_progs->progs)
	{
		if (mandatory)
			Host_Error ("PR_LoadProgs: couldn't load %s", filename);
		else
		{
			// this one is just a developer print so that we don't confuse the user; the rest will assume that csprogs loads and will print for real
			Con_DPrintf ("PR_LoadProgs: couldn't load %s\n", filename);
			return false;
		}
	}

	Con_DPrintf ("Programs occupy %iK.\n", com_filesize / 1024);

	for (i = 0; i < com_filesize; i++)
		CRC_ProcessByte (&pr_progs->crc, ((byte *) pr_progs->progs)[i]);

	// byte swap the header
	for (i = 0; i < sizeof (prvm_t) / 4; i++)
		((int *) pr_progs->progs)[i] = LittleLong (((int *) pr_progs->progs)[i]);

	if (pr_progs->progs->version != PROG_VERSION)
	{
		if (mandatory)
			Host_Error ("%s has wrong version number (%i should be %i)", filename, pr_progs->progs->version, PROG_VERSION);
		else
		{
			Con_Warning ("%s has wrong version number (%i should be %i)\n", filename, pr_progs->progs->version, PROG_VERSION);
			return false;
		}
	}

	// optionally send a 0 crc to allow for csqc crazy arsed behaviour (allowing multiple dufferent crcs indeed!)
	if (crc != 0 && pr_progs->progs->crc != crc)
	{
		if (mandatory)
			Host_Error ("%s system vars have been modified, progdefs.h is out of date", filename);
		else
		{
			Con_Warning ("%s system vars have been modified (CRC %i not %i)\nprogdefs.h is out of date\n", filename, pr_progs->progs->crc, crc);
			return false;
		}
	}

	pr_progs->functions = (dfunction_t *) ((byte *) pr_progs->progs + pr_progs->progs->ofs_functions);
	pr_progs->strings = (char *) pr_progs->progs + pr_progs->progs->ofs_strings;

	if (pr_progs->progs->ofs_strings + pr_progs->progs->numstrings >= com_filesize)
	{
		if (mandatory)
			Host_Error ("%s strings go past end of file", filename);
		else
		{
			Con_Warning ("%s strings go past end of file\n", filename);
			return false;
		}
	}

	pr_progs->numknownstrings = 0;
	pr_progs->maxknownstrings = 0;
	pr_progs->stringsize = pr_progs->progs->numstrings;

	pr_progs->knownstrings = NULL;
	PR_SetString ("");	// initialize the strings
	pr_progs->globaldefs = (ddef_t *) ((byte *) pr_progs->progs + pr_progs->progs->ofs_globaldefs);
	pr_progs->fielddefs = (ddef_t *) ((byte *) pr_progs->progs + pr_progs->progs->ofs_fielddefs);
	pr_progs->statements = (dstatement_t *) ((byte *) pr_progs->progs + pr_progs->progs->ofs_statements);

	pr_progs->global_struct = (globalvars_t *) ((byte *) pr_progs->progs + pr_progs->progs->ofs_globals);
	pr_progs->globals = (float *) pr_progs->global_struct;

	// byte swap the lumps
	for (i = 0; i < pr_progs->progs->numstatements; i++)
	{
		pr_progs->statements[i].op = LittleShort (pr_progs->statements[i].op);
		pr_progs->statements[i].a = LittleShort (pr_progs->statements[i].a);
		pr_progs->statements[i].b = LittleShort (pr_progs->statements[i].b);
		pr_progs->statements[i].c = LittleShort (pr_progs->statements[i].c);
	}

	// initialize function numbers for PROGS.DAT
	pr_progs->numbuiltins = 0;
	pr_progs->builtins = NULL;

	if (pr_builtin_remap.value)
	{
		// remove all previous assigned function numbers
		for (j = 1; j < pr_progs->ebfs_numbuiltins; j++)
			pr_progs->ebfs_builtins[j].funcno = 0;
	}
	else
	{
		// use default function numbers
		for (j = 1; j < pr_progs->ebfs_numbuiltins; j++)
		{
			pr_progs->ebfs_builtins[j].funcno = pr_progs->ebfs_builtins[j].default_funcno;

			// determine highest builtin number (when NOT remapped)
			if (pr_progs->ebfs_builtins[j].funcno > pr_progs->numbuiltins)
				pr_progs->numbuiltins = pr_progs->ebfs_builtins[j].funcno;
		}
	}

	for (i = 0; i < pr_progs->progs->numfunctions; i++)
	{
		pr_progs->functions[i].first_statement = LittleLong (pr_progs->functions[i].first_statement);
		pr_progs->functions[i].parm_start = LittleLong (pr_progs->functions[i].parm_start);
		pr_progs->functions[i].s_name = LittleLong (pr_progs->functions[i].s_name);
		pr_progs->functions[i].s_file = LittleLong (pr_progs->functions[i].s_file);
		pr_progs->functions[i].numparms = LittleLong (pr_progs->functions[i].numparms);
		pr_progs->functions[i].locals = LittleLong (pr_progs->functions[i].locals);

		if (pr_builtin_remap.value)
		{
			if (pr_progs->functions[i].first_statement < 0)	// builtin function
			{
				funcno = -pr_progs->functions[i].first_statement;
				funcname = PR_GetString (pr_progs->functions[i].s_name);

				// search function name
				for (j = 1; j < pr_progs->ebfs_numbuiltins; j++)
					if (!(Q_strcasecmp (funcname, pr_progs->ebfs_builtins[j].funcname)))
						break;	// found

				if (j < pr_progs->ebfs_numbuiltins)	// found
					pr_progs->ebfs_builtins[j].funcno = funcno;
				else Con_DPrintf ("Can not assign builtin number #%i to %s - function unknown\n", funcno, funcname);
			}
		}
	}

	if (pr_builtin_remap.value)
	{
		// check for unassigned functions and try to assign their default function number
		for (i = 1; i < pr_progs->ebfs_numbuiltins; i++)
		{
			if ((!pr_progs->ebfs_builtins[i].funcno) && (pr_progs->ebfs_builtins[i].default_funcno))	// unassigned and has a default number
			{
				// check if default number is already assigned to another function
				for (j = 1; j < pr_progs->ebfs_numbuiltins; j++)
					if (pr_progs->ebfs_builtins[j].funcno == pr_progs->ebfs_builtins[i].default_funcno)
						break;	// number already assigned to another builtin function

				if (j < pr_progs->ebfs_numbuiltins)	// already assigned
				{
					Con_DPrintf("Can not assign default builtin number #%i to %s - number is already assigned to %s\n",
					pr_progs->ebfs_builtins[i].default_funcno, pr_progs->ebfs_builtins[i].funcname, pr_progs->ebfs_builtins[j].funcname);
				}
				else pr_progs->ebfs_builtins[i].funcno = pr_progs->ebfs_builtins[i].default_funcno;
			}

			// determine highest builtin number (when remapped)
			if (pr_progs->ebfs_builtins[i].funcno > pr_progs->numbuiltins) pr_progs->numbuiltins = pr_progs->ebfs_builtins[i].funcno;
		}
	}

	pr_progs->numbuiltins++;

	// allocate and initialize builtin list for execution time
	pr_progs->builtins = Hunk_AllocName (pr_progs->numbuiltins * sizeof (builtin_t), "builtins");

	for (i = 0; i < pr_progs->numbuiltins; i++)
		pr_progs->builtins[i] = pr_progs->ebfs_builtins[0].function;

	// create builtin list for execution time and set cvars accordingly
	Cvar_Set ("pr_builtin_find", "0");
	Cvar_Set ("pr_checkextension", "0");	// 2001-10-20 Extension System by Lord Havoc/Maddes (DP compatibility)

	for (j = 1; j < pr_progs->ebfs_numbuiltins; j++)
	{
		// only put assigned functions into builtin list
		if (pr_progs->ebfs_builtins[j].funcno) pr_progs->builtins[pr_progs->ebfs_builtins[j].funcno] = pr_progs->ebfs_builtins[j].function;
		if (pr_progs->ebfs_builtins[j].default_funcno == PR_DEFAULT_FUNCNO_BUILTIN_FIND) Cvar_SetValue ("pr_builtin_find", pr_progs->ebfs_builtins[j].funcno);
		if (pr_progs->ebfs_builtins[j].default_funcno == PR_DEFAULT_FUNCNO_EXTENSION_FIND) Cvar_SetValue ("pr_checkextension", pr_progs->ebfs_builtins[j].funcno);
	}

	for (i = 0; i < pr_progs->progs->numglobaldefs; i++)
	{
		pr_progs->globaldefs[i].type = LittleShort (pr_progs->globaldefs[i].type);
		pr_progs->globaldefs[i].ofs = LittleShort (pr_progs->globaldefs[i].ofs);
		pr_progs->globaldefs[i].s_name = LittleLong (pr_progs->globaldefs[i].s_name);
	}

	for (i = 0; i < pr_progs->progs->numfielddefs; i++)
	{
		pr_progs->fielddefs[i].type = LittleShort (pr_progs->fielddefs[i].type);

		if (pr_progs->fielddefs[i].type & DEF_SAVEGLOBAL)
		{
			if (mandatory)
				Host_Error ("PR_LoadProgs: pr_progs->fielddefs[i].type & DEF_SAVEGLOBAL");
			else
			{
				Con_Warning ("PR_LoadProgs: pr_progs->fielddefs[i].type & DEF_SAVEGLOBAL\n");
				return false;
			}
		}

		pr_progs->fielddefs[i].ofs = LittleShort (pr_progs->fielddefs[i].ofs);
		pr_progs->fielddefs[i].s_name = LittleLong (pr_progs->fielddefs[i].s_name);
	}

	for (i = 0; i < pr_progs->progs->numglobals; i++)
		((int *) pr_progs->globals)[i] = LittleLong (((int *) pr_progs->globals)[i]);

	pr_progs->edict_size = pr_progs->progs->entityfields * 4 + sizeof (edict_t) - sizeof (entvars_t);

	// round to a power of 2 so that we can optimize divisions better
	for (i = 1, pr_progs->edict_shift = 0; i < pr_progs->edict_size; i <<= 1, pr_progs->edict_shift++);
	pr_progs->edict_size = i;

	FindEdictFieldOffsets ();

	end = SDL_GetTicks ();

	Con_DPrintf ("loaded progs in %i ms\n", (end - start));

	// round off to next highest whole word address (esp for Alpha)
	// this ensures that pointers in the engine data area are always
	// properly aligned
// 	pr_progs->edict_size += sizeof (void *) - 1;
// 	pr_progs->edict_size &= ~ (sizeof (void *) - 1);

	// progs are loaded OK now
	return true;
}


// 2001-09-14 Enhanced BuiltIn Function System (EBFS) by Maddes  start
/*
=============
PR_BuiltInList_f

For debugging, prints all builtin functions with assigned and default number
=============
*/
void PR_BuiltInList_f (void)
{
	int		i;
	char	*partial;
	int		len;
	int		count;

	if (Cmd_Argc () > 1)
	{
		partial = Cmd_Argv (1);
		len = strlen (partial);
	}
	else
	{
		partial = NULL;
		len = 0;
	}

	count = 0;

	for (i = 1; i < pr_progs->ebfs_numbuiltins; i++)
	{
		if (partial && Q_strncasecmp (partial, pr_progs->ebfs_builtins[i].funcname, len))
			continue;

		count++;
		Con_Printf ("%i(%i): %s\n", pr_progs->ebfs_builtins[i].funcno, pr_progs->ebfs_builtins[i].default_funcno, pr_progs->ebfs_builtins[i].funcname);
	}

	Con_Printf ("------------\n");

	if (partial)
		Con_Printf ("%i beginning with \"%s\" out of ", count, partial);

	Con_Printf ("%i builtin functions\n", i);
}
// 2001-09-14 Enhanced BuiltIn Function System (EBFS) by Maddes  end


// 2001-10-20 Extension System by LordHavoc  start
void PR_Extension_List_f (void)
{
	int		i;
	char	*partial;
	int		len;
	int		count;

	if (Cmd_Argc () > 1)
	{
		partial = Cmd_Argv (1);
		len = strlen (partial);
	}
	else
	{
		partial = NULL;
		len = 0;
	}

	count = 0;

	for (i = 0;; i++)
	{
		if (!pr_extensions[i]) break;

		if (partial && Q_strncasecmp (partial, pr_extensions[i], len))
			continue;

		count++;
		Con_Printf ("%s\n", pr_extensions[i]);
	}

	Con_Printf ("------------\n");

	if (partial)
	{
		Con_Printf ("%i beginning with \"%s\" out of ", count, partial);
	}

	Con_Printf ("%i extensions\n", i);
}
// 2001-10-20 Extension System by LordHavoc  end


/*
===============
PR_Init
===============
*/
void PR_Init (void)
{
	extern cvar_t pr_showstuffcmd;

	Cmd_AddCommand ("edict", ED_PrintEdict_f);
	Cmd_AddCommand ("edicts", ED_PrintEdicts);
	Cmd_AddCommand ("edictcount", ED_Count);
	Cmd_AddCommand ("profile", PR_Profile_f);
	Cmd_AddCommand ("builtinlist", PR_BuiltInList_f);	// 2001-09-14 Enhanced BuiltIn Function System (EBFS) by Maddes
	Cmd_AddCommand ("extensionlist", PR_Extension_List_f);	// 2001-10-20 Extension System by LordHavoc

	Cvar_RegisterVariable (&pr_showstuffcmd, NULL);
	Cvar_RegisterVariable (&nomonsters, NULL);
	Cvar_RegisterVariable (&gamecfg, NULL);
	Cvar_RegisterVariable (&scratch1, NULL);
	Cvar_RegisterVariable (&scratch2, NULL);
	Cvar_RegisterVariable (&scratch3, NULL);
	Cvar_RegisterVariable (&scratch4, NULL);
	Cvar_RegisterVariable (&savedgamecfg, NULL);
	Cvar_RegisterVariable (&saved1, NULL);
	Cvar_RegisterVariable (&saved2, NULL);
	Cvar_RegisterVariable (&saved3, NULL);
	Cvar_RegisterVariable (&saved4, NULL);

// 2001-09-14 Enhanced BuiltIn Function System (EBFS) by Maddes  start
	Cvar_RegisterVariable (&pr_builtin_find, NULL);
	Cvar_RegisterVariable (&pr_builtin_remap, NULL);
// 2001-09-14 Enhanced BuiltIn Function System (EBFS) by Maddes  end

	Cvar_RegisterVariable (&pr_checkextension, NULL);	// 2001-10-20 Extension System by LordHavoc (DP compatibility)
}


// ===========================================================================

#define	PR_STRING_ALLOCSLOTS	256

static void PR_AllocStringSlots (void)
{
	int oldmax = pr_progs->maxknownstrings;

	pr_progs->maxknownstrings += PR_STRING_ALLOCSLOTS;
	Con_DPrintf ("PR_AllocStringSlots: realloc'ing for %d slots\n", pr_progs->maxknownstrings);

	// don't abuse zone memory
	if (pr_progs->knownstrings)
	{
		char **tmpks = (char **) calloc (pr_progs->maxknownstrings * sizeof (char *), 1);
		memset (tmpks, 0, pr_progs->maxknownstrings * sizeof (char *));

		memcpy (tmpks, pr_progs->knownstrings, oldmax * sizeof (char *));

		free (pr_progs->knownstrings);
		pr_progs->knownstrings = tmpks;
	}
	else
	{
		pr_progs->knownstrings = (char **) calloc (pr_progs->maxknownstrings * sizeof (char *), 1);
		memset (pr_progs->knownstrings, 0, pr_progs->maxknownstrings * sizeof (char *));
	}
}

char *PR_GetString (int num)
{
	if (num >= 0 && num < pr_progs->stringsize)
		return pr_progs->strings + num;
	else if (num < 0 && num >= -pr_progs->numknownstrings)
	{
		if (!pr_progs->knownstrings[-1 - num])
		{
			Host_Error ("PR_GetString: attempt to get a non-existant string %d\n", num);
			return "";
		}

		return pr_progs->knownstrings[-1 - num];
	}
	else
	{
		Host_Error ("PR_GetString: invalid string offset %d\n", num);
		return "";
	}
}


int PR_SetString (char *s)
{
	int		i;

	if (!s)
		return 0;

	if (s >= pr_progs->strings && s <= pr_progs->strings + pr_progs->stringsize - 2)
		return (int) (s - pr_progs->strings);

	for (i = 0; i < pr_progs->numknownstrings; i++)
	{
		if (pr_progs->knownstrings[i] == s)
			return -1 - i;
	}

	// new unknown engine string
	if (i >= pr_progs->maxknownstrings)
		PR_AllocStringSlots ();

	pr_progs->numknownstrings++;
	pr_progs->knownstrings[i] = s;

	return -1 - i;
}

edict_t *GetEdictForNumber (int n)
{
	// edict overflow
	if (n >= MAX_EDICTS)
		Host_Error ("GetEdictForNumber: bad number %i (max: %i)", n, MAX_EDICTS);

	// do it this way because we're not guaranteed that 16 allocs will get us to n edicts
	while (n >= pr_progs->max_edicts)
	{
		if (pr_progs->max_edicts >= MAX_EDICTS)
			Host_Error ("GetEdictForNumber: edict overflow\n");

		// this can happen in some mods
		SV_AllocEdicts (16);
	}

	if (n < 0 || n >= pr_progs->max_edicts)
		Host_Error ("GetEdictForNumber: bad number %i (max: %i)", n, pr_progs->max_edicts);

	return pr_progs->edicts[n];
}

int GetNumberForEdict (edict_t *e)
{
	return e->ednum;
}


void PR_SetActiveProgs (prvm_t *progs)
{
	if (!progs)
		Sys_Error ("PR_SetActiveProgs : progs param is NULL");

	pr_progs = progs;
}


