/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// zone.c

#include "quakedef.h"


// for getting some big things off the hunk...
void *MallocZ (int size)
{
	void *data = malloc (size);
	memset (data, 0, size);

	return data;
}


// ============================================================================

#define	HUNK_SENTINAL	0x1df001ed

#define HUNKNAME_LEN	24
typedef struct
{
	int		sentinal;
	int		size;		// including sizeof(hunk_t), -1 = not allocated
	char	name[HUNKNAME_LEN];
} hunk_t;

byte	*hunk_base;
int		hunk_size;

int		hunk_low_used;
int		hunk_high_used;

qboolean	hunk_tempactive;
int		hunk_tempmark;


/*
==============
Hunk_Check

Run consistancy and sentinal trahing checks
==============
*/
void Hunk_Check (void)
{
	hunk_t	*h;

	for (h = (hunk_t *) hunk_base; (byte *) h != hunk_base + hunk_low_used ;)
	{
		if (h->sentinal != HUNK_SENTINAL)
			Sys_Error ("Hunk_Check: trahsed sentinal");

		if (h->size < sizeof (hunk_t) || h->size + (byte *) h - hunk_base > hunk_size)
			Sys_Error ("Hunk_Check: bad size");

		h = (hunk_t *) ((byte *) h + h->size);
	}
}

/*
==============
Hunk_Print

If "all" is specified, every single allocation is printed.
Otherwise, allocations with the same name will be totaled up before printing.
==============
*/
void Hunk_Print (qboolean all)
{
	hunk_t	*h, *next, *endlow, *starthigh, *endhigh;
	int		count, sum;
	int		totalblocks;
	char	name[9];

	name[8] = 0;
	count = 0;
	sum = 0;
	totalblocks = 0;

	h = (hunk_t *) hunk_base;
	endlow = (hunk_t *) (hunk_base + hunk_low_used);
	starthigh = (hunk_t *) (hunk_base + hunk_size - hunk_high_used);
	endhigh = (hunk_t *) (hunk_base + hunk_size);

	Con_Printf ("          :%8i total hunk size\n", hunk_size);
	Con_Printf ("-------------------------\n");

	while (1)
	{
		// 
		// skip to the high hunk if done with low hunk
		// 
		if (h == endlow)
		{
			Con_Printf ("-------------------------\n");
			Con_Printf ("          :%8i REMAINING\n", hunk_size - hunk_low_used - hunk_high_used);
			Con_Printf ("-------------------------\n");
			h = starthigh;
		}

		// 
		// if totally done, break
		// 
		if (h == endhigh)
			break;

		// 
		// run consistancy checks
		// 
		if (h->sentinal != HUNK_SENTINAL)
			Sys_Error ("Hunk_Check: trahsed sentinal");

		if (h->size < sizeof (hunk_t) || h->size + (byte *) h - hunk_base > hunk_size)
			Sys_Error ("Hunk_Check: bad size");

		next = (hunk_t *) ((byte *) h + h->size);
		count++;
		totalblocks++;
		sum += h->size;

		// 
		// print the single block
		// 
		memcpy (name, h->name, 8);

		if (all)
			Con_Printf ("%8p :%8i %8s\n", h, h->size, name);

		// 
		// print the total
		// 
		if (next == endlow || next == endhigh ||
				strncmp (h->name, next->name, 8))
		{
			if (!all)
				Con_Printf ("          :%8i %8s (TOTAL)\n", sum, name);

			count = 0;
			sum = 0;
		}

		h = next;
	}

	Con_Printf ("-------------------------\n");
	Con_Printf ("%8i total blocks\n", totalblocks);

}

/*
===================
Hunk_Print_f -- johnfitz -- console command to call hunk_print
===================
*/
void Hunk_Print_f (void)
{
	Hunk_Print (false);
}

/*
===================
Hunk_AllocName
===================
*/
void *Hunk_AllocName (int size, char *name)
{
	hunk_t	*h;

#ifdef PARANOID
	Hunk_Check ();
#endif

	if (size < 0)
		Sys_Error ("Hunk_Alloc: bad size: %i", size);

	size = sizeof (hunk_t) + ((size + 15) &~15);

	if (hunk_size - hunk_low_used - hunk_high_used < size)
		Sys_Error ("Hunk_Alloc: failed on %i bytes", size);

	h = (hunk_t *) (hunk_base + hunk_low_used);
	hunk_low_used += size;

	memset (h, 0, size);

	h->size = size;
	h->sentinal = HUNK_SENTINAL;
	Q_strncpy (h->name, name, HUNKNAME_LEN - 1);

	return (void *) (h + 1);
}

/*
===================
Hunk_Alloc
===================
*/
void *Hunk_Alloc (int size)
{
	return Hunk_AllocName (size, "unknown");
}

int	Hunk_LowMark (void)
{
	return hunk_low_used;
}

void Hunk_FreeToLowMark (int mark)
{
	if (mark < 0 || mark > hunk_low_used)
		Sys_Error ("Hunk_FreeToLowMark: bad mark %i", mark);

	memset (hunk_base + mark, 0, hunk_low_used - mark);
	hunk_low_used = mark;
}

int	Hunk_HighMark (void)
{
	if (hunk_tempactive)
	{
		hunk_tempactive = false;
		Hunk_FreeToHighMark (hunk_tempmark);
	}

	return hunk_high_used;
}

void Hunk_FreeToHighMark (int mark)
{
	if (hunk_tempactive)
	{
		hunk_tempactive = false;
		Hunk_FreeToHighMark (hunk_tempmark);
	}

	if (mark < 0 || mark > hunk_high_used)
		Sys_Error ("Hunk_FreeToHighMark: bad mark %i", mark);

	memset (hunk_base + hunk_size - hunk_high_used, 0, hunk_high_used - mark);
	hunk_high_used = mark;
}


/*
===================
Hunk_HighAllocName
===================
*/
void *Hunk_HighAllocName (int size, char *name)
{
	hunk_t	*h;

	if (size < 0)
		Sys_Error ("Hunk_HighAllocName: bad size: %i", size);

	if (hunk_tempactive)
	{
		Hunk_FreeToHighMark (hunk_tempmark);
		hunk_tempactive = false;
	}

#ifdef PARANOID
	Hunk_Check ();
#endif

	size = sizeof (hunk_t) + ((size + 15) &~15);

	if (hunk_size - hunk_low_used - hunk_high_used < size)
	{
		Con_Printf ("Hunk_HighAlloc: failed on %i bytes\n", size);
		return NULL;
	}

	hunk_high_used += size;

	h = (hunk_t *) (hunk_base + hunk_size - hunk_high_used);

	memset (h, 0, size);
	h->size = size;
	h->sentinal = HUNK_SENTINAL;
	Q_strncpy (h->name, name, sizeof (h->name) - 1);

	return (void *) (h + 1);
}


/*
=================
Hunk_TempAlloc

Return space from the top of the hunk
=================
*/
void *Hunk_TempAlloc (int size)
{
	void	*buf;

	size = (size + 15) &~15;

	if (hunk_tempactive)
	{
		Hunk_FreeToHighMark (hunk_tempmark);
		hunk_tempactive = false;
	}

	hunk_tempmark = Hunk_HighMark ();

	buf = Hunk_HighAllocName (size, "temp");

	hunk_tempactive = true;

	return buf;
}

/*
===============================================================================

CACHE MEMORY

===============================================================================
*/

// mh - new caching system to relieve pressure on fixed/bounded hunk memory
// fully dynamically allocated and no upper limit; works correctly with Fitz's texture freeing hack
typedef struct cache_system_s
{
	int						size;
	cache_user_t			*user;
	char					name[MAX_OSPATH];
	struct cache_system_s	*next;
	struct cache_system_s	*prev;
} cache_system_t;

cache_system_t *cache_head = NULL;
cache_system_t *cache_tail = NULL;


void Cache_FreeLow (int new_low_hunk) {/* used by hunk */}
void Cache_FreeHigh (int new_high_hunk) {/* used by Hunk */}
void Cache_Report (void) {/* used in cl_main.c */}
void *Cache_Check (cache_user_t *c) {return c->data;}
void Cache_Init (void) {/* used by init code */}

void Cache_Free (cache_user_t *c, qboolean freetextures)
{
	if (c)
	{
		cache_system_t *cs = ((cache_system_t *) c) - 1;

		if (freetextures && c && c->data)
			TM_FreeTexturesForOwner ((model_t *) (c + 1) - 1);

		if (cs->prev)
			cs->prev->next = cs->next;
		else cache_head = cs->next;

		if (cs->next)
			cs->next->prev = cs->prev;
		cache_tail = cs->prev;

		// prevent Cache_Check from blowing up
		cs->user->data = NULL;

		free (cs);
	}
}


void Cache_Flush (void)
{
	cache_system_t *cs = NULL;

	if (sv.active)
	{
		Con_Printf ("the cache cannot be flushed while the server is active!\n");
		return;
	}

	// because we're throwing stuff away we don't need to worry about relinking list pointers
	for (cs = cache_head; cs;)
	{
		cache_system_t *tmp = cs->next;

		if (cs->user)
		{
			if (cs->user->data)
				TM_FreeTexturesForOwner ((model_t *) (cs->user + 1) - 1);

			// prevent Cache_Check from blowing up
			cs->user->data = NULL;
		}

		free (cs);
		cs = tmp;
	}

	// !!! important !!!
	cache_head = cache_tail = NULL;
}


void *Cache_Alloc (cache_user_t *c, int size, char *name)
{
	cache_system_t *cs = NULL;

	for (cs = cache_head; cs; cs = cs->next)
	{
		if (!strcmp (cs->name, name))
		{
			Con_DPrintf ("retrieved %s from cache\n", cs->name);
			return cs->user->data;
		}
	}

	if (c->data) Sys_Error ("Cache_Alloc: allready allocated");
	if (size <= 0) Sys_Error ("Cache_Alloc: size %i", size);

	cs = (cache_system_t *) malloc (sizeof (cache_system_t) + size);
	cs->next = cache_head;

	if (!cache_head)
	{
		cache_head = cs;
		cs->prev = NULL;
	}
	else
	{
		cache_tail->next = cs;
		cs->prev = cache_tail;
	}

	cache_tail = cs;
	cs->next = NULL;

	strcpy (cs->name, name);
	cs->size = size;
	cs->user = c;

	c->data = (cs + 1);

	return c->data;
}


// ============================================================================


void Hunk_Size_f (void)
{
	Con_Printf ("%0.3f MB hunk used\n", ((float) Hunk_LowMark ()) / 1024.0f / 1024.0f);
}


/*
========================
Memory_Init
========================
*/
void Memory_Init (void *buf, int size)
{
	hunk_base = (byte *) buf;
	hunk_size = size;
	hunk_low_used = 0;
	hunk_high_used = 0;

	Cache_Init ();

	Cmd_AddCommand ("hunk_print", Hunk_Print_f); // johnfitz
	Cmd_AddCommand ("hunk_size", Hunk_Size_f); // mh
}

